/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2015 Grigori Goronzy <goronzy@kinoho.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "chachapoly.h"

/**
 * Constant-time memory compare. This should help with side-channel attacks.
 *
 * \param av input 1
 * \param bv input 2
 * \param n bytes to compare
 * \return 0 if inputs are equal
 */
static int memcmp_eq(const void *av, const void *bv, int n)
{
    const unsigned char *a = (const unsigned char*) av;
    const unsigned char *b = (const unsigned char*) bv;
    unsigned char res = 0;
    int i;

    for (i = 0; i < n; i++) {
        res |= *a ^ *b;
        a++;
        b++;
    }

    return res;
}

/**
 * poly1305 tag generation. This concatenates a string according to the rules
 * outlined in draft-agl-tls-chacha20poly1305-04 and calculates the tag.
 *
 * \param poly_key 32 byte secret one-time key for poly1305
 * \param ad associated data
 * \param ad_len associated data length in bytes
 * \param ct ciphertext
 * \param ct_len ciphertext length in bytes
 * \param tag pointer to 16 bytes for tag storage
 */
static void poly1305_get_tag(unsigned char *poly_key, const void *ad,
        int ad_len, const void *ct, int ct_len, unsigned char *tag)
{
    struct poly1305_context poly;
    uint64_t len;

    poly1305_init(&poly, poly_key);

    /* associated data and its length */
    len = ad_len;
    poly1305_update(&poly, ad, ad_len);
    poly1305_update(&poly, (unsigned char *)&len, 8);

    /* payload and its length */
    len = ct_len;
    poly1305_update(&poly, ct, ct_len);
    poly1305_update(&poly, (unsigned char *)&len, 8);

    poly1305_finish(&poly, tag);
}

/**
 * Initialize chacha20-poly1305 AEAD.
 *
 * \param ctx context data
 * \param key 16 or 32 bytes of key material
 * \param key_len key length, 256 or 512 bits
 * \return success if 0
 */
int chachapoly_init(struct chachapoly_ctx *ctx, const void *key, int key_len)
{
    memset(ctx, 0, sizeof(*ctx));
    chacha_keysetup(&ctx->cha_ctx, key, key_len);
    return 0;
}

/**
 * Encrypt or decrypt with chacha20-poly1305. The AEAD construction conforms
 * to draft-agl-tls-chacha20poly1305-04.
 *
 * \param ctx context data
 * \param nonce nonce (8 bytes)
 * \param ad associated data
 * \param ad_len associated data length in bytes
 * \param input plaintext/ciphertext input
 * \param input_len input length in bytes;
 * \param output plaintext/ciphertext output
 * \param tag tag output
 * \param tag_len tag length in bytes (0-16);
          if 0, authentification is skipped
 * \param encrypt decrypt if 0, else encrypt
 * \return CHACHAPOLY_OK if no error, CHACHAPOLY_INVALID_MAC if auth
 *         failed when decrypting
 */
int chachapoly_crypt(struct chachapoly_ctx *ctx, const void *nonce,
        const void *ad, int ad_len, void *input, int input_len,
        void *output, void *tag, int tag_len, int encrypt)
{
    unsigned char poly_key[CHACHA_BLOCKLEN];
    unsigned char calc_tag[POLY1305_TAGLEN];
    const unsigned char one[8] = { 1, 0, 0, 0, 0, 0, 0, 0 };

    /* initialize keystream and generate poly1305 key */
    memset(poly_key, 0, sizeof(poly_key));
    chacha_ivsetup(&ctx->cha_ctx, nonce, NULL);
    chacha_encrypt_bytes(&ctx->cha_ctx, poly_key, poly_key, sizeof(poly_key));

    /* check tag if decrypting */
    if (encrypt == 0 && tag_len) {
        poly1305_get_tag(poly_key, ad, ad_len, input, input_len, calc_tag);
        if (memcmp_eq(calc_tag, tag, tag_len) != 0) {
            return CHACHAPOLY_INVALID_MAC;
        }
    }

    /* crypt data */
    chacha_ivsetup(&ctx->cha_ctx, nonce, one);
    chacha_encrypt_bytes(&ctx->cha_ctx, (unsigned char *)input,
                         (unsigned char *)output, input_len);

    /* add tag if encrypting */
    if (encrypt && tag_len) {
        poly1305_get_tag(poly_key, ad, ad_len, output, input_len, calc_tag);
        memcpy(tag, calc_tag, tag_len);
    }

    return CHACHAPOLY_OK;
}

/**
 * Encrypt or decrypt with chacha20-poly1305 for short messages.
 * The AEAD construction is different from chachapoly_crypt, but more
 * efficient for small messages. Up to 32 bytes can be encrypted. The size
 * of associated data is not restricted. The interface is similar to
 * chachapoly_crypt.
 */
int chachapoly_crypt_short(struct chachapoly_ctx *ctx, const void *nonce,
        const void *ad, int ad_len, void *input, int input_len,
        void *output, void *tag, int tag_len, int encrypt)
{
    unsigned char keystream[CHACHA_BLOCKLEN];
    unsigned char calc_tag[POLY1305_TAGLEN];
    int i;

    assert(input_len <= 32);

    /* initialize keystream and generate poly1305 key */
    memset(keystream, 0, sizeof(keystream));
    chacha_ivsetup(&ctx->cha_ctx, nonce, NULL);
    chacha_encrypt_bytes(&ctx->cha_ctx, keystream, keystream,
            sizeof(keystream));

    /* check tag if decrypting */
    if (encrypt == 0 && tag_len) {
        poly1305_get_tag(keystream, ad, ad_len, input, input_len, calc_tag);
        if (memcmp_eq(calc_tag, tag, tag_len) != 0) {
            return CHACHAPOLY_INVALID_MAC;
        }
    }

    /* crypt data */
    for (i = 0; i < input_len; i++) {
        ((unsigned char *)output)[i] =
            ((unsigned char *)input)[i] ^ keystream[32 + i];
    }

    /* add tag if encrypting */
    if (encrypt && tag_len) {
        poly1305_get_tag(keystream, ad, ad_len, output, input_len, calc_tag);
        memcpy(tag, calc_tag, tag_len);
    }

    return CHACHAPOLY_OK;
}

int chachapoly_test(void)
{
    /* AEAD test vector from draft-agl-tls-chacha20poly1305-04 */
    unsigned char key[32] = {
        0x42, 0x90, 0xbc, 0xb1, 0x54, 0x17, 0x35, 0x31, 0xf3, 0x14, 0xaf,
        0x57, 0xf3, 0xbe, 0x3b, 0x50, 0x06, 0xda, 0x37, 0x1e, 0xce, 0x27,
        0x2a, 0xfa, 0x1b, 0x5d, 0xbd, 0xd1, 0x10, 0x0a, 0x10, 0x07
    };
    unsigned char ad[10] = {
        0x87, 0xe2, 0x29, 0xd4, 0x50, 0x08, 0x45, 0xa0, 0x79, 0xc0
    };
    unsigned char pt[10] = {
        0x86, 0xd0, 0x99, 0x74, 0x84, 0x0b, 0xde, 0xd2, 0xa5, 0xca
    };
    unsigned char nonce[8] = {
        0xcd, 0x7c, 0xf6, 0x7b, 0xe3, 0x9c, 0x79, 0x4a
    };
    unsigned char tag[16];
    int i, ret;
    struct chachapoly_ctx ctx;

    chachapoly_init(&ctx, key, 256);
    chachapoly_crypt(&ctx, nonce, ad, 10, pt, 10, pt, tag, 16, 1);

    printf("encrypt ct: ");
    for (i = 0; i < 10; i++)
        printf("%02X", pt[i]);
    printf(" tag: ");
    for (i = 0; i < 16; i++)
        printf("%02X", tag[i]);
    printf("\r\n");

    ret = chachapoly_crypt(&ctx, nonce, ad, 10, pt, 10, pt, tag, 16, 0);
    printf("decrypt status: %d\r\n", ret);

    return ret;
}

